clear;
clc;
close all;
warning off;

%% training data generator
N = 800;
n = 2;
x1 = randn(n,N/2)-0.8*ones(n,N/2);
y1 = zeros(N/2,1);
x2 = randn(n,N/2)+0.8*ones(n,N/2);
y2 = ones(N/2,1);
x = [x1,x2];
y = [y1;y2];
save train_data.mat x y N n
% load train_data.mat;

% To compute the testing accuracy, you may need to generate the testing
% data independently.

%% learning optimal classifier [w,b] using batch gradient descent
disp('Batch gradient descent training started...');
[w_batch,b_batch,cost_batch,w_buff_batch,b_buff_batch] = ...
    logistic_batch_train(x,y,1e-15,8000);
disp('Batch gradient descent training ended');

%% learning optimal classifier [w,b] using stochastic gradient descent
disp('Stochastic gradient descent training started...');
[w_sto,b_sto,cost_sto,w_buff_sto,b_buff_sto] = ...
    logistic_stochastic_train(x,y,1e-15,4000);
disp('Stochastic gradient descent training ended');

%% parameter & initialization
w = zeros(n,1);
b = zeros(1,1);
step = 0.001;
partial_w = zeros(n,1);
partial_b = 0;
cost_out = [];
epoch_out = 1;
count = 0;
max_iter_out = 4000;
alpha_buff = [];
k = 1;
lambda = 0.00005;
al_num = 1;
cost_out = logistic_cost(x,y,w,b);
cvx_sign = 0; % whether to use CVX toolbox

%% omniscient teaching
disp('Omniscient teaching started...');
while 1
    fprintf('Runing the %d-th iteration.\n',k);
    
%% choose the example
    for i=1:N
        a = 1/exp(w'*x(:,i)+b);
        partial_w(:,i) = (1/(a+1)-y(i))*x(:,i) + lambda*w;
        partial_b(:,i) = (1/(a+1)-y(i));
    end
    alpha = ones(N,1)/N;
    P = [partial_w;partial_b]'*[partial_w;partial_b]; 
    q = [partial_w;partial_b]' * ([w;b]-[w_batch;b_batch]);
    if cvx_sign == 1
        % CVX is not very useful in this simple demo, but it will be useful
        % in some other teaching scenarios.
        cvx_begin quiet
            cvx_precision high
            variable alpha(N)
            minimize( (step^2 * (diag(P))'  - 2*step*q')*alpha ) 
            subject to
                ones(1,N) * alpha == 1
                eye(N,N) * alpha >= zeros(N,1)
        cvx_end
        [~,alpha_rank] = sort(alpha,'descend');
        alpha_lib(al_num) = alpha_rank(1);
        select_idx = alpha_rank(1);
    else
        [~,select_idx] = min(step^2 * (diag(P))'  - 2*step*q');
        alpha_lib(al_num) = select_idx;
    end
    al_num = al_num+1;
    
%% update the classifier [w,b]
    w_out_buff(:,k) = w;
    b_out_buff(k) = b;
    k = k+1;
    a = 1/exp(w'*x(:,select_idx)+b);
    partial_w_out = (1/(a+1)-y(select_idx))*x(:,select_idx) + lambda*w;
    partial_b_out = (1/(a+1)-y(select_idx));
    w1 = w - step*partial_w_out;
    b1 = b - step*partial_b_out;
    new_cost_out = logistic_cost(x,y,w1,b1);
    cost_out = [cost_out,new_cost_out];
    w = w1;
    b = b1;
    count = count + 1;
    if count > max_iter_out
        break;
    end
    alpha_buff = [alpha_buff,alpha]; % learned alpha history
end
disp('Omniscient teaching ended');

%% plot results
figure1 = figure;
axes1 = axes('Parent',figure1);
hold(axes1,'on');
plot(cost_sto,'DisplayName','SGD','LineWidth',1);
plot(cost_out,'DisplayName','Omniscient Teaching','LineWidth',1);
xlim(axes1,[0 4000]);
box(axes1,'on');
legend1 = legend(axes1,'show');
clear;



